@@ -1,5 +1,6 @@
 from typing import Literal, Any
 from dataclasses import dataclass
+from datetime import datetime
 from anchorpy import Program
 from solana.publickey import PublicKey
 
@@ -29,6 +30,13 @@ class CollateralInfo:
     vault: PublicKey
 
 
+@dataclass(frozen=True)
+class FundingInfo:
+    hourly: float
+    daily: float
+    apr: float
+
+
 @dataclass(frozen=True)
 class MarketInfo:
     address: PublicKey
@@ -43,20 +51,39 @@ class MarketInfo:
     base_imf: int
     liq_fee: int
     index_price: float
-    index_twap: float
     mark_price: float
-    mark_twap: float
-    funding_rate: float
+    funding_sample_start_time: datetime
+    funding_info: None | FundingInfo
+
+    @property
+    def funding_rate(self):
+        import warnings
+
+        warnings.warn(
+            "Use of deprecated `MarketInfo.funding_rate`, please use `MarketInfo.funding_info`",
+            DeprecationWarning,
+        )
+        return 0 if self.funding_info is None else self.funding_info.hourly
 
 
 @dataclass(frozen=True)
 class PositionInfo:
     size: float
-    value: float
+    entry_value: float
     realized_pnl: float
     funding_index: float
     side: Literal["long", "short"]
 
+    @property
+    def value(self):
+        import warnings
+
+        warnings.warn(
+            "Use of deprecated `PositionInfo.value`, please use `PositionInfo.entry_value`",
+            DeprecationWarning,
+        )
+        return self.entry_value
+
 
 def order_type_from_str(t: OrderType, /, *, program: Program):
     typ = program.type["OrderType"]
