@@ -2,11 +2,13 @@ from typing import *
 import asyncio
 import os
 import json
-from datetime import datetime, timezone
+import inspect
+from datetime import datetime, timezone as tz
 from anchorpy import Idl, Program, Provider, Context, Wallet
 from anchorpy.error import AccountDoesNotExistError
 from solana.publickey import PublicKey
 from solana.keypair import Keypair
+from solana.transaction import TransactionInstruction, Transaction, TransactionSignature
 from solana.rpc.commitment import Commitment, Finalized
 from solana.rpc.async_api import AsyncClient
 from solana.rpc.types import TxOpts
@@ -17,12 +19,56 @@ from spl.token.constants import TOKEN_PROGRAM_ID
 
 from . import util, types, config
 from .config import configs, Config
-from .types import Side, OrderType, CollateralInfo, MarketInfo, PositionInfo
+from .types import (
+    Side,
+    OrderType,
+    CollateralInfo,
+    FundingInfo,
+    MarketInfo,
+    PositionInfo,
+)
 from .dex import Market, Orderbook, Order
 
 T = TypeVar("T")
 
 
+def GenIxDispatch(cls):
+    """Decorator for use with the `Zo` class.
+
+    This decorator finds all methods ending in `_ix` on the class,
+    and adds a trivial wrapper to dispatch the instruction using `Zo.send`.
+    The `__doc__` is moved to the generated method.
+    """
+    def gen(f):
+        async def g(self, *a, **kw):
+            # If you're looking for the source, see the source
+            # for the `_ix` variant of this method.
+            return await self.send(f(self, *a, **kw))
+        return g
+
+    for n, f in inspect.getmembers(cls, predicate=inspect.isfunction):
+        if n.startswith("_") or not n.endswith("_ix"):
+            continue
+
+        name = n[:-3]
+        g = gen(f)
+        g.__name__ = name
+        g.__qualname__ = f.__qualname__[:-3]
+        g.__doc__ = inspect.getdoc(f)
+        g.__signature__ = inspect.signature(f).replace(
+            return_annotation=TransactionSignature
+        )
+        g.__annotations__ = inspect.get_annotations(f)  # Copies.
+        g.__annotations__["return"] = TransactionSignature
+
+        # The docs are intended for the non '_ix' variant, so modify them.
+        f.__doc__ = f"See `Zo.{name}`."
+
+        setattr(cls, name, g)
+
+    return cls
+
+
 class ZoIndexer(Generic[T]):
     def __init__(self, d: dict[str, T], m: Callable[[str | int | PublicKey], str]):
         self.d = d
@@ -41,6 +87,7 @@ class ZoIndexer(Generic[T]):
         return self.d[self.m(i)]
 
 
+@GenIxDispatch
 class Zo:
     __program: Program
     __config: Config
@@ -49,13 +96,13 @@ class Zo:
     __collaterals: dict[str, CollateralInfo]
     __orderbook: dict[str, Orderbook]
     __balance: dict[str, float]
-    __position: dict[str, float]
+    __position: dict[str, PositionInfo]
 
     __dex_markets: dict[str, Market]
     __orders: dict[str, list[Order]]
 
-    __markets_map: dict[str | int, str]
-    __collaterals_map: dict[str | int, str]
+    __markets_map: dict[str | int | PublicKey, str]
+    __collaterals_map: dict[str | int | PublicKey, str]
 
     _zo_state: Any
     _zo_state_signer: PublicKey
@@ -189,53 +236,35 @@ class Zo:
     def wallet(self) -> Wallet:
         return self.provider.wallet
 
-    def _collaterals_map(self, k: str | int | PublicKey) -> str:
-        if isinstance(k, PublicKey):
-            for i, c in enumerate(self._zo_state.collaterals):
-                if c.mint == k:
-                    return self.__collaterals_map[i]
-            raise ValueError("")
-        else:
-            return self.__collaterals_map[k]
-
-    def _markets_map(self, k: str | int | PublicKey) -> str:
-        if isinstance(k, PublicKey):
-            for i, m in enumerate(self._zo_state.perp_markets):
-                if m.dex_market == k:
-                    return self.__markets_map[i]
-            raise ValueError("")
-        else:
-            return self.__markets_map[k]
-
     @property
     def collaterals(self):
         """List of collaterals and their metadata."""
-        return ZoIndexer(self.__collaterals, lambda k: self._collaterals_map(k))
+        return ZoIndexer(self.__collaterals, lambda k: self.__collaterals_map[k])
 
     @property
     def markets(self):
         """List of collaterals and markets metadata."""
-        return ZoIndexer(self.__markets, lambda k: self._markets_map(k))
+        return ZoIndexer(self.__markets, lambda k: self.__markets_map[k])
 
     @property
     def orderbook(self):
         """Current state of the orderbook."""
-        return ZoIndexer(self.__orderbook, lambda k: self._markets_map(k))
+        return ZoIndexer(self.__orderbook, lambda k: self.__markets_map[k])
 
     @property
     def balance(self):
         """Current account balance."""
-        return ZoIndexer(self.__balance, lambda k: self._collaterals_map(k))
+        return ZoIndexer(self.__balance, lambda k: self.__collaterals_map[k])
 
     @property
     def position(self):
         """Current position."""
-        return ZoIndexer(self.__position, lambda k: self._markets_map(k))
+        return ZoIndexer(self.__position, lambda k: self.__markets_map[k])
 
     @property
     def orders(self):
         """Currently active orders."""
-        return ZoIndexer(self.__orders, lambda k: self._markets_map(k))
+        return ZoIndexer(self.__orders, lambda k: self.__markets_map[k])
 
     def _get_open_orders_info(self, key: int | str, /):
         if isinstance(key, str):
@@ -259,6 +288,7 @@ class Zo:
             symbol = util.decode_symbol(c.oracle_symbol)
             map[symbol] = symbol
             map[i] = symbol
+            map[c.mint] = symbol
 
             collaterals[symbol] = CollateralInfo(
                 mint=c.mint,
@@ -290,8 +320,13 @@ class Zo:
                 break
 
             symbol = util.decode_symbol(m.symbol)
+
+            if symbol == "LUNA-PERP":
+                continue
+
             map[symbol] = symbol
             map[i] = symbol
+            map[m.dex_market] = symbol
 
             oracle = None
             for o in reversed(self._zo_cache.oracles):
@@ -302,34 +337,27 @@ class Zo:
                 raise IndexError(f"oracle for market {symbol} not found")
 
             mark = self._zo_cache.marks[i]
-            twap = self._zo_cache.marks[i].twap
 
             price_adj = 10 ** (m.asset_decimals - 6)
             index_price = util.decode_wrapped_i80f48(oracle.price) * price_adj
-            index_twap = util.decode_wrapped_i80f48(oracle.twap) * price_adj
             mark_price = util.decode_wrapped_i80f48(mark.price) * price_adj
 
             if types.perp_type_to_str(m.perp_type, program=self.program) == "square":
                 index_price = index_price**2 / m.strike
-                index_twap = index_twap**2 / m.strike
 
-            twap_sample_since_intervals = (
-                datetime.fromtimestamp(
-                    twap.last_sample_start_time, tz=timezone.utc
-                ).minute
-                // 5
+            funding_sample_start = datetime.fromtimestamp(
+                mark.twap.last_sample_start_time, tz=tz.utc
             )
-            mark_twap = (
-                mark_price
-                if twap_sample_since_intervals == 0
-                else (
-                    util.decode_wrapped_i80f48(twap.cumul_avg)
-                    * price_adj
-                    / (4 * twap_sample_since_intervals)
+            cumul_avg = util.decode_wrapped_i80f48(mark.twap.cumul_avg)
+            if abs(cumul_avg) == 0 or funding_sample_start.minute == 0:
+                funding_info = None
+            else:
+                daily_funding = cumul_avg / funding_sample_start.minute
+                funding_info = FundingInfo(
+                    daily=daily_funding,
+                    hourly=daily_funding / 24,
+                    apr=daily_funding * 100 * 365,
                 )
-            )
-
-            funding_rate = min(0.05, max(-0.05, mark_twap / index_twap - 1)) * 100 / 24
 
             markets[symbol] = MarketInfo(
                 address=m.dex_market,
@@ -344,10 +372,9 @@ class Zo:
                 base_imf=m.base_imf,
                 liq_fee=m.liq_fee,
                 index_price=index_price,
-                index_twap=index_twap,
                 mark_price=mark_price,
-                mark_twap=mark_twap,
-                funding_rate=funding_rate,
+                funding_sample_start_time=funding_sample_start,
+                funding_info=funding_info,
             )
 
         self.__markets_map = map
@@ -385,7 +412,7 @@ class Zo:
                     size=util.small_to_big_amount(
                         abs(oo.pos_size), decimals=m.base_decimals
                     ),
-                    value=util.small_to_big_amount(
+                    entry_value=util.small_to_big_amount(
                         abs(oo.native_pc_total), decimals=m.quote_decimals
                     ),
                     realized_pnl=util.small_to_big_amount(
@@ -398,7 +425,7 @@ class Zo:
                 )
             else:
                 positions[s] = PositionInfo(
-                    size=0, value=0, realized_pnl=0, funding_index=1, side="long"
+                    size=0, entry_value=0, realized_pnl=0, funding_index=1, side="long"
                 )
 
         self.__position = positions
@@ -408,21 +435,23 @@ class Zo:
         ks = [
             m.dex_market
             for m in self._zo_state.perp_markets
-            if m.dex_market != PublicKey(0)
+            if m.dex_market in self.__markets_map
         ]
         res: Any = await self.connection.get_multiple_accounts(
             ks, encoding="base64", commitment=commitment
         )
         res = res["result"]["value"]
         self.__dex_markets = {
-            self.__markets_map[i]: Market.from_base64(res[i]["data"][0])
-            for i in range(len(self.__markets))
+            self.__markets_map[ks[i]]: Market.from_base64(res[i]["data"][0])
+            for i in range(len(ks))
         }
 
-    async def __reload_orders(self, *, commitment: None | Commitment = None):
+    async def refresh_orders(self, *, commitment: None | Commitment = None):
+        """Refresh only the orderbook and orders."""
+        ss = []
         ks = []
-        for i in range(len(self.__markets)):
-            mkt = self.__dex_markets[self.__markets_map[i]]
+        for s, mkt in self.__dex_markets.items():
+            ss.append(s)
             ks.extend((mkt.bids, mkt.asks))
 
         res: Any = await self.connection.get_multiple_accounts(
@@ -432,12 +461,12 @@ class Zo:
         orders = self._zo_margin and {}
         orderbook = {}
 
-        for i in range(len(self.__markets)):
-            mkt = self.__dex_markets[self.__markets_map[i]]
+        for i, s in enumerate(ss):
+            mkt = self.__dex_markets[s]
             ob = mkt._decode_orderbook_from_base64(
                 res[2 * i]["data"][0], res[2 * i + 1]["data"][0]
             )
-            orderbook[self.__markets_map[i]] = ob
+            orderbook[s] = ob
 
             if self._zo_margin is not None:
                 os = []
@@ -445,7 +474,7 @@ class Zo:
                     for o in slab:
                         if o.control == self._zo_margin.control:
                             os.append(o)
-                orders[self.__markets_map[i]] = os
+                orders[s] = os
 
         self.__orderbook = orderbook
         self.__orders = orders
@@ -472,9 +501,25 @@ class Zo:
         self.__reload_balances()
         self.__reload_positions()
         await self.__reload_dex_markets(commitment=commitment)
-        await self.__reload_orders(commitment=commitment)
+        await self.refresh_orders(commitment=commitment)
 
-    async def deposit(
+    async def send(
+        self, *xs: TransactionInstruction | Transaction
+    ) -> TransactionSignature:
+        """Dispatch a set of instructions.
+
+        Args:
+            xs: The instruction or transaction to add.
+
+        Returns:
+            The transaction signature.
+        """
+        tx = Transaction()
+        for x in xs:
+            tx.add(x)
+        return await self.provider.send(tx)
+
+    def deposit_ix(
         self,
         amount: float,
         /,
@@ -482,7 +527,7 @@ class Zo:
         mint: PublicKey,
         repay_only: bool = False,
         token_account: None | PublicKey = None,
-    ) -> str:
+    ) -> TransactionInstruction:
         """Deposit collateral into the margin account.
 
         Args:
@@ -501,7 +546,7 @@ class Zo:
         decimals = self.collaterals[mint].decimals
         amount = util.big_to_small_amount(amount, decimals=decimals)
 
-        return await self.program.rpc["deposit"](
+        return self.program.instruction["deposit"](
             repay_only,
             amount,
             ctx=Context(
@@ -518,7 +563,7 @@ class Zo:
             ),
         )
 
-    async def withdraw(
+    def withdraw_ix(
         self,
         amount: float,
         /,
@@ -526,7 +571,7 @@ class Zo:
         mint: PublicKey,
         allow_borrow: bool = False,
         token_account: None | PublicKey = None,
-    ) -> str:
+    ) -> TransactionInstruction:
         """Withdraw collateral from the margin account.
 
         Args:
@@ -548,7 +593,7 @@ class Zo:
         decimals = self.collaterals[mint].decimals
         amount = util.big_to_small_amount(amount, decimals=decimals)
 
-        return await self.program.rpc["withdraw"](
+        return self.program.instruction["withdraw"](
             allow_borrow,
             amount,
             ctx=Context(
@@ -566,7 +611,7 @@ class Zo:
             ),
         )
 
-    async def place_order(
+    def place_order_ix(
         self,
         size: float,
         price: float,
@@ -576,7 +621,8 @@ class Zo:
         order_type: OrderType,
         limit: int = 10,
         client_id: int = 0,
-    ) -> str:
+        max_ts: None | int = None,
+    ) -> Transaction:
         """Place an order on the orderbook.
 
         Args:
@@ -594,6 +640,8 @@ class Zo:
                 be used to cancel this order through
                 cancelPerpOrderByClientId. For optimal use, make sure
                 all ids for every order is unique.
+            max_ts: If the current on-chain Unix timestamp exceeds this
+                value, then the order will not go through.
 
         Returns:
             The transaction signature.
@@ -645,12 +693,11 @@ class Zo:
                             "rent": SYSVAR_RENT_PUBKEY,
                             "system_program": SYS_PROGRAM_ID,
                         },
-                        pre_instructions=pre_ixs,
                     )
                 )
             ]
 
-        return await self.program.rpc["place_perp_order"](
+        return self.program.instruction["place_perp_order_with_max_ts"](
             is_long,
             price,
             base_qty,
@@ -658,6 +705,7 @@ class Zo:
             order_type_,
             limit,
             client_id,
+            max_ts if max_ts is not None else 2**63 - 1,
             ctx=Context(
                 accounts={
                     "state": self.__config.ZO_STATE_ID,
@@ -674,25 +722,26 @@ class Zo:
                     "market_asks": mkt.asks,
                     "dex_program": self.__config.ZO_DEX_ID,
                     "rent": SYSVAR_RENT_PUBKEY,
-                }
+                },
+                pre_instructions=pre_ixs,
             ),
         )
 
-    async def __cancel_order(
+    def __cancel_order_ix(
         self,
         *,
         symbol: str,
         side: None | Side = None,
         order_id: None | int = None,
         client_id: None | int = None,
-    ):
+    ) -> TransactionInstruction:
         mkt = self.__dex_markets[symbol]
         oo = self._get_open_orders_info(symbol)
 
         if oo is None:
             raise ValueError("open orders account is uninitialized")
 
-        return await self.program.rpc["cancel_perp_order"](
+        return self.program.instruction["cancel_perp_order"](
             order_id,
             side == "bid",
             client_id,
@@ -713,9 +762,9 @@ class Zo:
             ),
         )
 
-    async def cancel_order_by_order_id(
+    def cancel_order_by_order_id_ix(
         self, order_id: int, side: Side, *, symbol: str
-    ) -> str:
+    ) -> TransactionInstruction:
         """Cancel an order on the orderbook using the `order_id`.
 
         Args:
@@ -727,9 +776,11 @@ class Zo:
         Returns:
             The transaction signature.
         """
-        return await self.__cancel_order(symbol=symbol, order_id=order_id, side=side)
+        return self.__cancel_order_ix(symbol=symbol, order_id=order_id, side=side)
 
-    async def cancel_order_by_client_id(self, client_id: int, *, symbol: str) -> str:
+    def cancel_order_by_client_id_ix(
+        self, client_id: int, *, symbol: str
+    ) -> TransactionInstruction:
         """Cancel an order on the orderbook using the `client_id`.
 
         Args:
@@ -740,4 +791,27 @@ class Zo:
         Returns:
             The transaction signature.
         """
-        return await self.__cancel_order(symbol=symbol, client_id=client_id)
+        return self.__cancel_order_ix(symbol=symbol, client_id=client_id)
+
+    def close_position_ix(self, symbol: str, /, *, max_ts: int | None = None):
+        """Close a position.
+
+        Args:
+            symbol: The markey symbol, e.g. "BTC-PERP".
+            max_ts: See `Zo.place_order`.
+
+        Returns:
+            The transaction signature.
+        """
+        pos = self.position[symbol]
+        side = "ask" if pos.side == "long" else "bid"
+        price = 0 if side == "ask" else 2**63 - 1
+
+        return self.place_order_ix(
+            pos.size,
+            price,
+            side,
+            symbol=symbol,
+            order_type="reduceonlyioc",
+            max_ts=max_ts,
+        )
